import sys
import os

# RTD changes directory into docs/ to run the build,
# so the module code cannot be found for autodoc
if os.environ.get('READTHEDOCS') == 'True':
  sys.path.append("../src")
else:
  sys.path.append("./src")

project = "Relay.sh SDK"
author = "Puppet Inc"
copyright = f"2020, {author}"
extensions = [
  "sphinx.ext.autodoc",
  "sphinx.ext.napoleon",
  "sphinx_autodoc_typehints"
]



"The main class for a client to connect to the Relay service API"
from __future__ import annotations

import json
from typing import Any, Optional, Union

from .client import new_session
from .events import Events
from .outputs import Outputs
from .util import json_object_hook


class UnresolvableException(Exception):
    pass


class DynamicMetaclass(type):

    def __getattr__(self, name: str) -> Dynamic:
        return Dynamic(name)


class Dynamic(metaclass=DynamicMetaclass):
    """A query interface for inspecting a spec.

    This class allows arbitrary traversal that can be converted to a query to
    the metadata API.
    """

    def __init__(self, name: str, parent: Optional[Dynamic] = None) -> None:
        self._name = name
        self._parent = parent

    def __getattr__(self, name: str) -> Dynamic:
        return Dynamic(name, parent=self)

    def __str__(self) -> str:
        if self._parent is None:
            return self._name

        return '{0}[{1}]'.format(self._parent, json.dumps(self._name))


class Interface:
    """An Interface object connects client code to the metadata service."""

    def __init__(self, api_url: Optional[str] = None):
        self._client = new_session(api_url=api_url)

    def get(self, q: Optional[Union[Dynamic, str]] = None) -> Any:
        """Retrieve values from the metadata service

        Args:
            q: A particular parameter to query the value of.

        Returns:
            The value of the queried parameter as a string.
            If no query was provided, all available parameters will be
            returned in a json map
        """

        params = {}
        if q is not None:
            params['q'] = str(q)

        r = self._client.get('http+api://api/spec', params=params)
        r.raise_for_status()

        data = json.loads(r.text, object_hook=json_object_hook)
        if not data['complete']:
            raise UnresolvableException()

        return data['value']

    @property
    def events(self) -> Events:
        """Accessor for Events methods"""
        return Events(self._client)

    @property
    def outputs(self) -> Outputs:
        """Accessor for Outputs methods"""
        return Outputs(self._client)



import asyncio
from concurrent import futures
from typing import Any, Dict, Iterable

import pytest


def event_loop_exception_handler(loop: asyncio.AbstractEventLoop,
                                 context: Dict[str, Any]) -> None:
    # Ignore dropped tasks in test.
    if loop.is_closed() and isinstance(context.get('task'), asyncio.Task):
        return

    loop.default_exception_handler(context)


@pytest.fixture
def event_loop() -> Iterable[asyncio.AbstractEventLoop]:
    loop = asyncio.new_event_loop()
    loop.set_exception_handler(event_loop_exception_handler)
    yield loop
    loop.close()


@pytest.fixture
def thread_pool_executor() -> Iterable[futures.Executor]:
    with futures.ThreadPoolExecutor() as pool:
        yield pool



"""A SDK for interacting with the relay.sh infrastructure services"""
from .interface import Dynamic, Interface, UnresolvableException
from .util import (NoTerminationPolicy, SignalTerminationPolicy,
                   SoftTerminationPolicy, TerminationPolicy)
from .webhook import WebhookServer

__all__ = [
    'Dynamic',
    'Interface',
    'UnresolvableException',
    'NoTerminationPolicy',
    'SignalTerminationPolicy',
    'SoftTerminationPolicy',
    'TerminationPolicy',
    'WebhookServer',
]



"""Local data types"""
from typing import Container, Text, Tuple, Union

HTTPTimeout = Union[float, Tuple[float, float], Tuple[float, None]]
HTTPClientCertificate = Union[bytes, Text, Container[Union[bytes, Text]]]



import asyncio
import threading
from concurrent import futures
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Mapping, Union

import pytest
from hypercorn.typing import ASGIFramework
from quart import Quart
from relay_sdk.util import SoftTerminationPolicy
from relay_sdk.webhook import WebhookServer
from requests import Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

if TYPE_CHECKING:
    from wsgiref.types import StartResponse, WSGIApplication


retry_strategy = Retry(
    total=4,
    backoff_factor=1,
)

session = Session()
session.mount('http://', HTTPAdapter(max_retries=retry_strategy))


class TestWebhookServer:

    async def _test_app(self, event_loop: asyncio.AbstractEventLoop,
                        app: Union[ASGIFramework, 'WSGIApplication']) -> None:
        # Create the server.
        term = SoftTerminationPolicy()

        srv = WebhookServer(app, termination_policy=term, port=0)
        srv_task = event_loop.create_task(srv.serve())

        # Issue request to server and check response.
        resp = await event_loop.run_in_executor(
            None, session.get,
            f'http://localhost:{srv.port}',
        )
        resp.raise_for_status()

        assert resp.json() == {'success': True}

        # Stop the server.
        term.terminate_task(srv_task)

        # Wait for everything to clean up and exit.
        await asyncio.wait_for(asyncio.gather(*filter(
            lambda t: t != asyncio.current_task(),
            asyncio.all_tasks(),
        ), return_exceptions=True), 30)

    @pytest.mark.asyncio
    async def test_asgi_2(self, event_loop: asyncio.AbstractEventLoop) -> None:
        class Application:

            def __init__(self, scope: Dict[Any, Any]) -> None:
                if scope['type'] != 'http':
                    raise NotImplementedError()

            async def __call__(self, receive: Callable[..., Any],
                               send: Callable[..., Any]) -> None:
                await send({
                    'type': 'http.response.start',
                    'status': 200,
                })
                await send({
                    'type': 'http.response.body',
                    'body': b'{"success": true}'
                })

        await self._test_app(event_loop, Application)

    @pytest.mark.asyncio
    async def test_asgi_3(self, event_loop: asyncio.AbstractEventLoop) -> None:
        async def application(scope: Dict[Any, Any],
                              receive: Callable[..., Any],
                              send: Callable[..., Any]) -> None:
            if scope['type'] != 'http':
                raise NotImplementedError()

            await send({
                'type': 'http.response.start',
                'status': 200,
            })
            await send({
                'type': 'http.response.body',
                'body': b'{"success": true}'
            })

        await self._test_app(event_loop, application)

    @pytest.mark.asyncio
    async def test_wsgi(self, event_loop: asyncio.AbstractEventLoop) -> None:
        def application(environ: Mapping[str, Any],
                        start_response: 'StartResponse') -> Iterable[bytes]:
            start_response('200 OK', [])
            yield b'{"success": true}'

        await self._test_app(event_loop, application)

    @pytest.mark.asyncio
    async def test_quart(self, event_loop: asyncio.AbstractEventLoop) -> None:
        application = Quart(__name__)

        @application.route('/')
        async def hello() -> str:
            return '{"success": true}'

        await self._test_app(event_loop, application)

    def test_serve_forever(self,
                           thread_pool_executor: futures.Executor) -> None:
        """This test ensures that the serve_forever() method runs correctly
        and will allow outstanding connections to clean up cleanly before
        exiting.
        """

        req_ev, term_ev = threading.Event(), threading.Event()

        # This application waits for a termination event on the main thread
        # before it allows itself to complete the request.
        def application(environ: Mapping[str, Any],
                        start_response: 'StartResponse') -> Iterable[bytes]:
            start_response('200 OK', [])

            req_ev.set()
            term_ev.wait()

            yield b'OK'

        term = SoftTerminationPolicy()

        # Run the server ("forever") in another thread.
        srv = WebhookServer(application, termination_policy=term, port=0)
        srv_t = thread_pool_executor.submit(srv.serve_forever)

        def request() -> None:
            resp = session.get(f'http://localhost:{srv.port}')
            resp.raise_for_status()

            assert resp.text == 'OK'

        # Open a request to the server in a new thread. The request will block
        # until we send our termination signal.
        req_t = thread_pool_executor.submit(request)

        # Wait for us to reach the request handler. After this the request will
        # be held open.
        req_ev.wait(timeout=30)

        # Request termination of server.
        term.terminate_all()

        # Wait for listener to close.
        while srv.listening():
            pass

        # Allow the request to complete.
        term_ev.set()

        # Make sure everything gets cleaned up. This will propagate assertion
        # errors from each thread.
        srv_t.result(timeout=30)
        req_t.result(timeout=30)



"Generating events for the service to act on"
import json
from typing import Any, Mapping

from requests import Session

from .util import JSONEncoder


class Events:
    """Class for generating event payloads to the service API"""

    def __init__(self, client: Session) -> None:
        self._client = client

    def emit(self, data: Mapping[str, Any]) -> None:
        """Create and send an event to the service.

        Use this from a Trigger handler to start the workflow
        associated with the trigger.

        Accepts a Mapping of data fields that will bind
        names of workflow parameters to values from the
        request payload that came in to the Trigger handler"""
        r = self._client.post(
            'http+api://api/events',
            data=json.dumps({'data': data}, cls=JSONEncoder),
            headers={'content-type': 'application/json'},
        )
        r.raise_for_status()



"""A helper class for building Trigger webhooks for Relay"""
from __future__ import annotations

import asyncio
import os
from typing import TYPE_CHECKING, Optional, Union, cast

from asgiref.wsgi import WsgiToAsgi
from hypercorn.asyncio.run import worker_serve
from hypercorn.config import Config, Sockets
from hypercorn.typing import ASGIFramework

from .util import SignalTerminationPolicy, TerminationPolicy, is_async_callable

if TYPE_CHECKING:
    from wsgiref.types import WSGIApplication

DEFAULT_TERMINATION_POLICY = SignalTerminationPolicy(timeout_sec=25)
DEFAULT_PORT = 8080


class PortParseError(Exception):
    pass


class WebhookServer:
    """Instantiate a WebhookServer class to handle incoming webhook payloads

    They should use the serve_forever() startup method, because Relay
    will manage the lifecycle of the trigger container.

    They should not expect any state to be preserved between requests,
    nor rely on any persistence on the local container filesystem.
    """

    _app: ASGIFramework
    _config: Config
    _sockets: Sockets
    _termination_policy: TerminationPolicy
    _task: Optional[asyncio.Task[None]]

    def __init__(self, app: Union[ASGIFramework, 'WSGIApplication'], *,
                 termination_policy: Optional[TerminationPolicy] = None,
                 port: Optional[int] = None):
        if not is_async_callable(app):
            app = WsgiToAsgi(app)

        if termination_policy is None:
            termination_policy = DEFAULT_TERMINATION_POLICY

        if port is None:
            try:
                port = int(os.environ['PORT'])
            except KeyError:
                port = DEFAULT_PORT
            except ValueError:
                raise PortParseError()

        self._app = cast(ASGIFramework, app)
        self._config = Config()
        self._config.bind = [f'0.0.0.0:{port}']
        self._sockets = self._config.create_sockets()
        self._termination_policy = termination_policy
        self._task = None

    @property
    def port(self) -> int:
        return self._sockets.insecure_sockets[0].getsockname()[1]

    def listening(self) -> bool:
        return self._task is not None and not self._task.done()

    async def serve(self) -> None:
        if self._task is None:
            loop = asyncio.get_running_loop()
            shutdown_trigger = await self._termination_policy.attach()

            self._task = loop.create_task(worker_serve(
                self._app, self._config,
                sockets=self._sockets,
                shutdown_trigger=shutdown_trigger,
            ))

        await self._task

    async def _serve_wait(self) -> None:
        current_task = asyncio.current_task()
        assert current_task is not None

        await self.serve()
        await asyncio.gather(
            *filter(lambda t: t != current_task, asyncio.all_tasks()),
            return_exceptions=True,
        )

    def serve_forever(self) -> None:
        asyncio.run(self._serve_wait())



"""Internal class for clients to interact with the metadata service"""
import os
from typing import Mapping, Optional, Union
from urllib.parse import urljoin, urlsplit, urlunsplit

from requests import PreparedRequest, Response, Session
from requests.adapters import HTTPAdapter

from .typing import HTTPClientCertificate, HTTPTimeout


class MetadataAPIAdapter(HTTPAdapter):

    def __init__(self, base_url: Optional[str] = None) -> None:
        if base_url is None:
            base_url = os.environ['METADATA_API_URL']

        self._base_url = base_url

        super(MetadataAPIAdapter, self).__init__()

    def send(self, request: PreparedRequest, stream: bool = False,
             timeout: Optional[HTTPTimeout] = None,
             verify: Union[bool, str] = True,
             cert: Optional[HTTPClientCertificate] = None,
             proxies: Optional[Mapping[str, str]] = None) -> Response:
        (_, _, path, query, fragment) = urlsplit(request.url or '')
        """Sends a prepared http request to the metadata API server"""
        request.prepare_url(
            urljoin(
                self._base_url,
                urlunsplit(('', '', path, query, fragment)),
            ),
            {},
        )

        return super(MetadataAPIAdapter, self).send(
            request,
            stream=stream,
            timeout=timeout,
            verify=verify,
            cert=cert,
            proxies=proxies,
        )


def new_session(api_url: Optional[str] = None) -> Session:
    """Create a new client session to the metadata API.

    Args:
        api_url: host and port to communicate with the api server (optional)

    Returns:
        a Session object connected to the api server
    """
    sess = Session()
    sess.mount('http+api://', MetadataAPIAdapter(base_url=api_url))
    return sess
